//==================================================================================================
/*
  EVE - Expressive Vector Engine
  Copyright : EVE Project Contributors
  SPDX-License-Identifier: BSL-1.0
*/
//==================================================================================================
#pragma once

#include <eve/arch.hpp>
#include <eve/detail/overload.hpp>
#include <eve/module/core/regular/all.hpp>
#include <eve/module/core/regular/is_flint.hpp>

namespace eve
{
//================================================================================================
//! @addtogroup core_internal
//! @{
//!   @var ldexp
//!   @brief Computes \f$\textstyle x 2^n\f$.
//!
//!   **Defined in Header**
//!
//!   @code
//!   #include <eve/module/core.hpp>
//!   @endcode
//!
//!   @groupheader{Callable Signatures}
//!
//!   @code
//!   namespace eve
//!   {
//!      template< eve::value T, eve::integral_value N >
//!      eve::common_compatible_t<T, N> ldexp(T x, N n) noexcept;
//!   }
//!   @endcode
//!
//!   **Parameters**
//!
//!     * `x`:  [floating argument](@ref eve::floating_value).
//!
//!     * `n` :  [integral value argument](@ref eve::integral_value).
//!
//!    **Return value**
//!
//!    The value of \f$\textstyle x 2^n\f$ is returned.
//!
//!  @groupheader{Example}
//!
//!  @godbolt{doc/core/regular/ldexp.cpp}
//!
//!  @groupheader{Semantic Modifiers}
//!
//!   * Masked Call
//!
//!     The call `eve::ldexp[mask](x, ...)` provides a masked
//!     version of `ldexp` which is
//!     equivalent to `if_else(mask, ldexp(x, ...), x)`
//!
//!      **Example**
//!
//!        @godbolt{doc/core/masked/ldexp.cpp}
//!
//! @}
//================================================================================================
namespace tag
{
  struct ldexp_;
}

namespace detail
{
  template<typename T, typename U>
  EVE_FORCEINLINE void check(EVE_SUPPORTS(eve::tag::ldexp_), T const&, [[maybe_unused]] U const& b)
  {
    if constexpr( std::is_floating_point_v<element_type_t<U>> )
      EVE_ASSERT(eve::all(is_flint(b)), "[eve::ldexp] argument 2 is floating but not a flint");
  }
}

EVE_MAKE_CALLABLE(ldexp_, ldexp);
}

#include <eve/module/core/regular/impl/ldexp.hpp>

#if defined(EVE_INCLUDE_X86_HEADER)
#  include <eve/module/core/regular/impl/simd/x86/ldexp.hpp>
#endif
